@@ -4,67 +4,67 @@ Contains the different types of aggregator that can be used in the definition of
 a Measure in the Schema.
 """
 
-import abc
-import dataclasses
-from typing import Dict, List, Set
+import dataclasses as dcls
+from typing import Any, Dict, List, Tuple, Union
 
 from .enums import AggregatorType
 
 
-class Aggregator(abc.ABC):
+def _get_fields(obj: object) -> List[str]:
+    """Returns a set of strings, with the names of the fields in this class."""
+    return sorted(f.name for f in dcls.fields(obj)) if dcls.is_dataclass(obj) else []
+
+
+class Aggregator:
     """Base class for aggregator methods.
 
     The instances contain parameters to compose the associated SQL query.
     Backend packages should extend these according to support, and implement
     the methods needed for conversion to string.
     """
-    def __init__(self) -> None:
-        ...
 
-    def __repr__(self) -> str:
-        return "{0}({1})".format(self, ", ".join(
-            "{0}={1}".format(f.name, repr(getattr(self, f.name)))
-            for f in dataclasses.fields(self)
-        ) if dataclasses.is_dataclass(self) else "")
+    def __repr__(self):
+        params = ", ".join(
+            f"{field}={repr(getattr(self, field))}" for field in _get_fields(self)
+        )
+        return f"{type(self).__name__}({params})"
 
     def __str__(self) -> str:
         return type(self).__name__
 
-    @classmethod
-    def get_fields(cls) -> Set[str]:
-        """Returns a set of strings, with the names of the fields in this class.
-        """
-        return set(f.name for f in dataclasses.fields(cls)) \
-               if dataclasses.is_dataclass(cls) else \
-               set()
+    def __iter__(self):
+        yield "type", type(self).__name__
+        yield from ((field, getattr(self, field)) for field in _get_fields(self))
 
     def get_params(self) -> Dict[str, str]:
         """Returns a dict with the serialized params of the instance.
 
         Base class just returns an empty dict. More complex subclasses must
         override this method."""
-        return {}
+        return dict((field, getattr(self, field)) for field in _get_fields(self))
 
     @classmethod
-    def new(cls, kwargs: dict) -> "Aggregator":
-        field_names = cls.get_fields()
+    def new(cls, kwargs: Dict[str, Any]) -> "Aggregator":
+        field_names = _get_fields(cls)
         return cls(**{k: v for k, v in kwargs.items() if k in field_names})
 
     @staticmethod
     def from_enum(enum: AggregatorType) -> "Aggregator":
         agg_classes = {
-            AggregatorType.SUM: Sum,
-            AggregatorType.COUNT: Count,
             AggregatorType.AVERAGE: Average,
+            AggregatorType.COUNT: Count,
             AggregatorType.MAX: Max,
+            AggregatorType.MEDIAN: Median,
             AggregatorType.MIN: Min,
             AggregatorType.MODE: Mode,
+            AggregatorType.SUM: Sum,
             AggregatorType.BASICGROUPEDMEDIAN: BasicGroupedMedian,
-            AggregatorType.WEIGHTEDSUM: WeightedSum,
-            AggregatorType.WEIGHTEDAVERAGE: WeightedAverage,
-            AggregatorType.REPLICATEWEIGHTMOE: ReplicateWeightMoe,
             AggregatorType.CALCULATEDMOE: CalculatedMoe,
+            AggregatorType.QUANTILE: Quantile,
+            AggregatorType.REPLICATEWEIGHTMOE: ReplicateWeightMoe,
+            AggregatorType.WEIGHTEDAVERAGE: WeightedAverage,
             AggregatorType.WEIGHTEDAVERAGEMOE: WeightedAverageMoe,
+            AggregatorType.WEIGHTEDSUM: WeightedSum,
         }
         return agg_classes[enum]
 
@@ -93,16 +93,29 @@ class Mode(Aggregator):
     pass
 
 
-@dataclasses.dataclass
+class Median(Aggregator):
+    pass
+
+
+@dcls.dataclass(frozen=True)
+class Quantile(Aggregator):
+    """Quantile is calculated against the measure's value column.
+
+    `quantile(quantile_level)(column)`
+
+    where quantile_level = Quantile level between 0 and 1
+    """
+
+    quantile_level: Union[str, float]
+
+
+@dcls.dataclass(frozen=True)
 class BasicGroupedMedian(Aggregator):
     group_aggregator: str
     group_dimension: str
 
-    def get_params(self) -> Dict[str, str]:
-        return dataclasses.asdict(self)
 
-
-@dataclasses.dataclass
+@dcls.dataclass(frozen=True)
 class WeightedSum(Aggregator):
     """Weighted Sum is calculated against the measure's value column.
 
@@ -111,25 +124,21 @@ class WeightedSum(Aggregator):
     First roll-up is sum(column * weight_column) as weighted_sum_first
     Second roll-up is sum(weighted_sum_first) as weighted_sum_final
     """
+
     weight_column: str
 
-    def get_params(self) -> Dict[str, str]:
-        return dataclasses.asdict(self)
 
-
-@dataclasses.dataclass
+@dcls.dataclass(frozen=True)
 class WeightedAverage(Aggregator):
     """Weighted Average is calculated against the measure's value column.
 
     `sum(column * weight_column) / sum(weight_column)`
     """
-    weight_column: str
 
-    def get_params(self) -> Dict[str, str]:
-        return dataclasses.asdict(self)
+    weight_column: Union[str, Tuple[str, float]]
 
 
-@dataclasses.dataclass
+@dcls.dataclass(frozen=True)
 class ReplicateWeightMoe(Aggregator):
     """Where the measure column is the primary value, and a list of secondary
     columns is provided to the MO aggregator:
@@ -141,15 +150,13 @@ class ReplicateWeightMoe(Aggregator):
     where cv = critical value, for 90% confidence interval it's 1.645
     where df = design factor / #samples
     """
+
     critical_value: float
     design_factor: float
     secondary_columns: List[str]
 
-    def get_params(self) -> Dict[str, str]:
-        return dataclasses.asdict(self)
 
-
-@dataclasses.dataclass
+@dcls.dataclass(frozen=True)
 class CalculatedMoe(Aggregator):
     """Where the moe is already calculated for each row, and this just
     aggregates them correctly.
@@ -158,13 +165,11 @@ class CalculatedMoe(Aggregator):
 
     where cv = critical value; for 90% confidence interval it's 1.645
     """
-    critical_value: float
 
-    def get_params(self) -> Dict[str, str]:
-        return dataclasses.asdict(self)
+    critical_value: Union[str, float]
 
 
-@dataclasses.dataclass
+@dcls.dataclass(frozen=True)
 class WeightedAverageMoe(Aggregator):
     """
     Where the measure column is the primary value,
@@ -177,10 +182,8 @@ class WeightedAverageMoe(Aggregator):
     where cv = critical value, for 90% confidence interval it's 1.645
     where df = design factor / #samples
     """
+
     critical_value: float
     design_factor: float
     primary_weight: str
     secondary_weight_columns: List[str]
-
-    def get_params(self) -> Dict[str, str]:
-        return dataclasses.asdict(self)
