@@ -1,27 +1,69 @@
 import logging
+from collections import defaultdict
 from functools import lru_cache
 from itertools import chain
-from typing import (Generic, Iterable, Iterator, List, Mapping, Optional,
-                    OrderedDict, Set, TypeVar, Union)
+from typing import (
+    TYPE_CHECKING,
+    Dict,
+    Generic,
+    Iterable,
+    Iterator,
+    Mapping,
+    Optional,
+    OrderedDict,
+    Set,
+    TypeVar,
+    Union,
+)
+
+import immutables as immu
 
 from tesseract_olap.common import T, get_localization
-from tesseract_olap.query.exceptions import (InvalidEntityName,
-                                             TimeDimensionUnavailable,
-                                             TimeScaleUnavailable)
+from tesseract_olap.exceptions.query import (
+    InvalidEntityName,
+    TimeDimensionUnavailable,
+    TimeScaleUnavailable,
+)
+from tesseract_olap.exceptions.schema import (
+    DuplicatedNameError,
+    EntityUsageError,
+    InvalidNameError,
+    MissingPropertyError,
+)
 
 from .enums import DimensionType, MemberType
-from .exceptions import (DuplicatedNameError, EntityUsageError,
-                         MissingPropertyError)
-from .models import (Annotations, CaptionSet, Cube, Dimension, DimensionUsage,
-                     Entity, Hierarchy, HierarchyUsage, InlineTable, Level,
-                     LevelUsage, Measure, Property, PropertyUsage, Schema,
-                     Table, Usage)
+from .models import (
+    Annotations,
+    AnyMeasure,
+    CalculatedMeasure,
+    CaptionSet,
+    Cube,
+    Dimension,
+    DimensionUsage,
+    Entity,
+    Hierarchy,
+    HierarchyUsage,
+    InlineTable,
+    Level,
+    LevelUsage,
+    Measure,
+    Property,
+    PropertyUsage,
+    Schema,
+    Table,
+    Usage,
+)
+
+if TYPE_CHECKING:
+    from tesseract_olap.query import RequestWithRoles
 
 logger = logging.getLogger(__name__)
 
 EntityType = TypeVar("EntityType", bound=Entity)
 UsageType = TypeVar("UsageType", bound=Usage)
 
+ColumnEntity = Union["AnyMeasure", "LevelTraverser", "PropertyTraverser"]
+
 
 class SchemaTraverser(Mapping[str, "CubeTraverser"]):
     """Wrapper class for Schema model, to generate the relationships between the
@@ -31,9 +73,14 @@ class SchemaTraverser(Mapping[str, "CubeTraverser"]):
     def __init__(self, schema: "Schema"):
         self.schema = schema
         self.cube_map = OrderedDict(
-            (name, CubeTraverser(cube,
-                                 dimension_map=schema.shared_dimension_map,
-                                 table_map=schema.shared_table_map))
+            (
+                name,
+                CubeTraverser(
+                    cube,
+                    dimension_map=schema.shared_dimension_map,
+                    table_map=schema.shared_table_map,
+                ),
+            )
             for name, cube in schema.cube_map.items()
         )
 
@@ -56,46 +103,77 @@ class SchemaTraverser(Mapping[str, "CubeTraverser"]):
         except KeyError:
             raise InvalidEntityName("Cube", cube_name) from None
 
-    def get_locale_available(self) -> List[str]:
-        locales = set()
+    def get_locale_available(self) -> Set[str]:
+        locales = {self.schema.default_locale}
         for item in self.cube_map.values():
             locales.update(item.get_locale_available())
         # TODO: add from shared_dimension_map and shared_table_map
         locales.discard("xx")
-        locales.add(self.schema.default_locale)
-        return sorted(locales)
+        return locales
 
-    def get_public_schema(self, locale: Optional[str] = None):
-        """Generates a JSON-encodeable schema describing this entity."""
-        default_locale = self.schema.default_locale
-        locale = default_locale if locale is None else locale
-        return {
-            "name": self.schema.name,
-            # "locales": self.get_locale_available(),
-            "default_locale": default_locale,
-            "cubes": [cube.get_public_schema(locale)
-                      for cube in self.cube_map.values()],
-            "annotations": self.schema.annotations,
-        }
+    def is_authorized(self, request: "RequestWithRoles") -> bool:
+        """Validates a request has enough permissions to be executed."""
+        return self.get_cube(request.cube).is_authorized(request.roles)
 
     def validate(self):
         """Performs some verifications on the resulting data structure, after
         parsing the schema file."""
 
-        def validate_unique_names(nameset: Set[str], item: "EntityUsageTraverser"):
-            if item.name in nameset:
-                origin = item._entity if item._usage is None else item._usage
-                raise DuplicatedNameError(type(origin).__name__, item.name)
-            nameset.add(item.name)
-
         for cube in self.cube_map.values():
-            levelset = set()
-            propsset = set()
+            cubename = cube.name
+            nameset = {item.name: type(item).__name__ for item in cube.measures}
+
+            for measure in cube.measures:
+                for item in measure.and_submeasures():
+                    validate_entity_names(cubename, item, nameset)
 
             for level in cube.levels:
-                validate_unique_names(levelset, level)
+                validate_entity_names(cubename, level, nameset)
                 for prop in level.properties:
-                    validate_unique_names(propsset, prop)
+                    validate_entity_names(cubename, prop, nameset)
+
+    def unwrap_tables(self):
+        """Extracts the {table: column[]} data from all entities in the schema."""
+        tables: dict[str, set[str]] = defaultdict(set)
+
+        for cube in self.cube_map.values():
+            table = cube.table
+            if isinstance(table, InlineTable):
+                continue
+
+            # Index fact tables
+            tables[table.name].update(
+                [table.primary_key],
+                (
+                    item.key_column
+                    for measure in cube.measures
+                    for item in measure.and_submeasures()
+                    if isinstance(item, Measure)
+                ),
+                (dimension.foreign_key for dimension in cube.dimensions),
+            )
+
+            for hierarchy in cube.hierarchies:
+                table = hierarchy.table
+                if table is None or isinstance(table, InlineTable):
+                    continue
+
+                # Index dimension tables
+                tables[table.name].update(
+                    [table.primary_key],
+                    (
+                        item
+                        for level in hierarchy.levels
+                        for item in (level.key_column, *level.name_column_map.values())
+                    ),
+                    (
+                        item
+                        for propty in hierarchy.properties
+                        for item in propty.key_column_map.values()
+                    ),
+                )
+
+        return tables
 
 
 class CubeTraverser:
@@ -109,6 +187,12 @@ class CubeTraverser:
     _cube: "Cube"
     _dimension_map: Mapping[str, "DimensionTraverser"]
     _table: Union["Table", "InlineTable"]
+    annotations: Annotations
+    captions: CaptionSet
+    measure_map: Mapping[str, "AnyMeasure"]
+    name: str
+    subset_table: bool
+    visible: bool
 
     def __init__(
         self,
@@ -119,49 +203,56 @@ class CubeTraverser:
     ) -> None:
         self._cube = cube
         self._dimension_map = OrderedDict(
-            (name, DimensionTraverser(item, table_map=table_map)
-                   if isinstance(item, Dimension) else
-                   DimensionTraverser(
-                       get_shared_entity(dimension_map, item.source), item,
-                       table_map=table_map))
+            (
+                name,
+                DimensionTraverser(item, table_map=table_map)
+                if isinstance(item, Dimension)
+                else DimensionTraverser(
+                    get_shared_entity(dimension_map, item.source),
+                    item,
+                    table_map=table_map,
+                ),
+            )
             for name, item in cube.dimension_map.items()
         )
-        self._table = table_map[cube.table] \
-                      if isinstance(cube.table, str) else \
-                      cube.table
+        self.measure_map = immu.Map(
+            (item.name, item)
+            for measure in self._cube.measure_map.values()
+            for item in measure.and_submeasures()
+        )
+        self._table = (
+            table_map[cube.table] if isinstance(cube.table, str) else cube.table
+        )
 
-    def __repr__(self) -> str:
+    def __repr__(self):
         return f"CubeTraverser(name='{self._cube.name}', table={self.table})"
 
-    @property
-    def name(self) -> str:
-        return self._cube.name
-
-    @property
-    def annotations(self) -> Annotations:
-        return self._cube.annotations
-
-    @property
-    def captions(self) -> CaptionSet:
-        return self._cube.captions
+    def __getattr__(self, _name: str):
+        return getattr(self._cube, _name)
 
     @property
     def table(self) -> Union["Table", "InlineTable"]:
         return self._table
 
     @property
-    def measures(self) -> Iterable["Measure"]:
+    def measures(self) -> Iterable["AnyMeasure"]:
         return self._cube.measure_map.values()
 
+    @property
+    def calculated_measures(self) -> Iterable["CalculatedMeasure"]:
+        return (
+            item
+            for item in self._cube.measure_map.values()
+            if isinstance(item, CalculatedMeasure)
+        )
+
     @property
     def dimensions(self) -> Iterable["DimensionTraverser"]:
         return self._dimension_map.values()
 
     @property
     def time_dimensions(self) -> Iterable["DimensionTraverser"]:
-        return (item
-                for item in self.dimensions
-                if item.dim_type == DimensionType.TIME)
+        return (item for item in self.dimensions if item.dim_type == DimensionType.TIME)
 
     @property
     def hierarchies(self) -> Iterable["HierarchyTraverser"]:
@@ -188,7 +279,7 @@ class CubeTraverser:
     def get_caption(self, locale: str = "xx") -> str:
         return self._cube.get_caption(locale)
 
-    def get_locale_available(self) -> List[str]:
+    def get_locale_available(self) -> Set[str]:
         """Returns a list of strings containing the locale code keys available
         for captions in entities inside this :class:`Cube`."""
         locales = set(self.captions.keys())
@@ -196,15 +287,15 @@ class CubeTraverser:
             locales.update(item.get_locale_available())
         for item in self.measures:
             locales.update(item.get_locale_available())
-        return sorted(locales)
+        return locales
 
-    def get_measure(self, name: str) -> "Measure":
+    def get_measure(self, name: str) -> "AnyMeasure":
         """Attempts to retrieve a Measure by its name.
 
         Raises :class:`InvalidEntityName` if the entity can't be found.
         """
         try:
-            return self._cube.measure_map[name]
+            return self.measure_map[name]
         except KeyError:
             raise InvalidEntityName("Measure", name) from None
 
@@ -250,8 +341,9 @@ class CubeTraverser:
         dimension = None
 
         for dimension in self.time_dimensions:
-            exact_match = next((item for item in dimension.levels
-                               if item.name == scale), None)
+            exact_match = next(
+                (item for item in dimension.levels if item.name == scale), None
+            )
             if exact_match is not None:
                 return exact_match
 
@@ -259,8 +351,9 @@ class CubeTraverser:
             raise TimeDimensionUnavailable(self._cube.name)
 
         for dimension in self.time_dimensions:
-            substr_match = next((item for item in dimension.levels
-                                if scale in item.name.lower()), None)
+            substr_match = next(
+                (item for item in dimension.levels if scale in item.name.lower()), None
+            )
             if substr_match is not None:
                 return substr_match
 
@@ -276,37 +369,8 @@ class CubeTraverser:
         except StopIteration:
             raise InvalidEntityName("Property", name) from None
 
-    @lru_cache(maxsize=1)
-    def get_public_schema(self, locale: str):
-        """Generates a JSON-encodeable schema describing this entity."""
-        return {
-            "name": self.name,
-            "caption": self._cube.get_caption(locale),
-            "dimensions": [
-                item.get_public_schema(locale)
-                for item in self.dimensions
-            ],
-            "measures": [
-                {
-                    "name": measure.name,
-                    "caption": measure.get_caption(locale),
-                    "aggregator": {
-                        "name": str(measure.aggregator),
-                    },
-                    "attached": [
-                        {
-                            "name": submsr.name,
-                            "caption": submsr.get_caption(locale),
-                            "aggregator": str(submsr.aggregator),
-                        }
-                        for submsr in measure.submeasures.values()
-                    ],
-                    "annotations": measure.annotations,
-                }
-                for measure in self.measures
-            ],
-            "annotations": self._cube.annotations,
-        }
+    def is_authorized(self, roles: Iterable[str]) -> bool:
+        return self._cube.is_authorized(roles)
 
 
 class EntityUsageTraverser(Generic[EntityType, UsageType]):
@@ -324,7 +388,7 @@ class EntityUsageTraverser(Generic[EntityType, UsageType]):
         self._entity = entity
         self._usage = usage
 
-    def __contains__(self, item) -> bool:
+    def __contains__(self, item: Union[EntityType, UsageType]) -> bool:
         return item == self._entity or item == self._usage
 
     def __dir__(self) -> Iterable[str]:
@@ -333,8 +397,8 @@ class EntityUsageTraverser(Generic[EntityType, UsageType]):
     def __getattr__(self, name: str):
         return getattr(self._entity, name)
 
-    def __repr__(self) -> str:
-        return "{}(name='{}')".format(self.__class__.__name__, self.name)
+    def __repr__(self):
+        return f"{type(self).__name__}(name={repr(self.name)})"
 
     @property
     def name(self) -> str:
@@ -373,11 +437,11 @@ class EntityUsageTraverser(Generic[EntityType, UsageType]):
         caption = get_localization(self.captions, locale)
         return self.name if caption is None else caption
 
-    def get_locale_available(self) -> List[str]:
+    def get_locale_available(self) -> Set[str]:
         """Retrieves the list of locales for whose a caption has been defined in
         this entity.
         """
-        return sorted(self.captions.keys())
+        return set(self.captions.keys())
 
 
 class DimensionTraverser(EntityUsageTraverser[Dimension, DimensionUsage]):
@@ -402,9 +466,12 @@ class DimensionTraverser(EntityUsageTraverser[Dimension, DimensionUsage]):
             )
         else:
             self.hierarchy_map = OrderedDict(
-                (name, HierarchyTraverser(
-                    entity.hierarchy_map[item.source], item, table_map=table_map
-                ))
+                (
+                    name,
+                    HierarchyTraverser(
+                        entity.hierarchy_map[item.source], item, table_map=table_map
+                    ),
+                )
                 for name, item in usage.hierarchy_map.items()
             )
 
@@ -416,69 +483,64 @@ class DimensionTraverser(EntityUsageTraverser[Dimension, DimensionUsage]):
         # depends solely on self._usage presence.
         if self._usage is not None:
             return self._usage.foreign_key
-        if self._entity.foreign_key is None:  # should be unreachable, but checks types
+        if self._entity.foreign_key is None:
+            # should be unreachable, but checks types
             raise MissingPropertyError("Dimension", self.name, "foreign_key")
         return self._entity.foreign_key
 
     @property
-    def default_hierarchy(self):
-        hie = self._entity.default_hierarchy
+    def default_hierarchy(self) -> "HierarchyTraverser":
+        hie = self._entity.get_default_hierarchy()
         if self._usage is None:
             return self.hierarchy_map[hie.name]
         try:
+            # find the HierarchyTraverser that references the default Hierarchy
             return next(item for item in self.hierarchies if hie in item)
         except StopIteration:
             return next(iter(self.hierarchies))
 
     @property
     def hierarchies(self) -> Iterable["HierarchyTraverser"]:
-        """Returns a generator that yields all Hierarchies under this Cube."""
+        """Returns a generator that yields all Hierarchies under this Dimension."""
         return self.hierarchy_map.values()
 
     @property
     def levels(self) -> Iterable["LevelTraverser"]:
-        """Returns a generator that yields all Levels under this Cube."""
+        """Returns a generator that yields all Levels under this Dimension."""
         return chain(*(item.levels for item in self.hierarchies))
 
     @property
     def properties(self) -> Iterable["PropertyTraverser"]:
-        """Returns a generator that yields all Properties under this Cube."""
+        """Returns a generator that yields all Properties under this Dimension."""
         return chain(*(item.properties for item in self.hierarchies))
 
     def get_hierarchy(self, name: str) -> "HierarchyTraverser":
-        """Retrieves a Hierarchy from this Cube by its name."""
+        """Retrieves a Hierarchy from this Dimension by its name.
+        Raises :class:`InvalidEntityName` if the Hierarchy can't be found.
+        """
         try:
             return next(item for item in self.hierarchies if item.name == name)
         except StopIteration:
             raise InvalidEntityName("Hierarchy", name) from None
 
     def get_level(self, name: str) -> "LevelTraverser":
-        """Retrieves a Level from this Cube by its name."""
+        """Retrieves a Level from this Dimension by its name.
+        Raises :class:`InvalidEntityName` if the Level can't be found.
+        """
         try:
             return next(item for item in self.levels if item.name == name)
         except StopIteration:
             raise InvalidEntityName("Level", name) from None
 
     def get_property(self, name: str) -> "PropertyTraverser":
-        """Retrieves a Property from this Cube by its name."""
+        """Retrieves a Property from this Dimension by its name.
+        Raises :class:`InvalidEntityName` if the Property can't be found.
+        """
         try:
             return next(item for item in self.properties if item.name == name)
         except StopIteration:
             raise InvalidEntityName("Property", name) from None
 
-    def get_public_schema(self, locale: str):
-        """Generates a JSON-encodeable schema describing this entity."""
-        return {
-            "name": self.name,
-            "caption": self.get_caption(locale),
-            "type": self.dim_type.value,
-            "hierarchies": [
-                item.get_public_schema(locale) for item in self.hierarchies
-            ],
-            "default_hierarchy": self.default_hierarchy.name,
-            "annotations": self.annotations,
-        }
-
 
 class HierarchyTraverser(EntityUsageTraverser[Hierarchy, HierarchyUsage]):
     """Allows seamless aliasing of values between a Hierarchy and a HierarchyUsage."""
@@ -548,15 +610,6 @@ class HierarchyTraverser(EntityUsageTraverser[Hierarchy, HierarchyUsage]):
         except StopIteration:
             raise InvalidEntityName("Property", name) from None
 
-    def get_public_schema(self, locale: str):
-        """Generates a JSON-encodeable schema describing this entity."""
-        return {
-            "name": self.name,
-            "caption": self.get_caption(locale),
-            "levels": [item.get_public_schema(locale) for item in self.levels],
-            "annotations": self.annotations,
-        }
-
 
 class LevelTraverser(EntityUsageTraverser[Level, LevelUsage]):
     """Allows seamless aliasing of values between a Level and a LevelUsage."""
@@ -564,6 +617,7 @@ class LevelTraverser(EntityUsageTraverser[Level, LevelUsage]):
     depth: int
     key_column: str
     key_type: MemberType
+    name_column_map: Mapping[str, str]
     name: str
     property_map: Mapping[str, "PropertyTraverser"]
 
@@ -581,43 +635,52 @@ class LevelTraverser(EntityUsageTraverser[Level, LevelUsage]):
                 for name, item in usage.property_map.items()
             )
 
+    @property
+    def type_caster(self):
+        return self.key_type.get_caster()
+
     @property
     def properties(self) -> Iterable["PropertyTraverser"]:
         """Returns a generator that yields all Properties under this Level."""
         return self.property_map.values()
 
-    def get_name_column(self, locale: str = "xx"):
+    def get_name_column(self, locale: str = "xx") -> Union[str, None]:
         return self._entity.get_name_column(locale)
 
-    def get_public_schema(self, locale: str):
-        """Generates a JSON-encodeable schema describing this entity."""
-        return {
-            "name": self.name,
-            "caption": self.get_caption(locale),
-            "properties": [item.get_public_schema(locale) for item in self.properties],
-            "annotations": self.annotations,
-        }
-
 
 class PropertyTraverser(EntityUsageTraverser[Property, PropertyUsage]):
     """Allows seamless aliasing of values between a Property and a PropertyUsage."""
 
+    key_column_map: Mapping[str, str]
+    key_type: MemberType
+
     def get_key_column(self, locale: str = "xx"):
         """Returns the matching `key_column` of this Property for a certain locale."""
         return self._entity.get_key_column(locale)
 
-    def get_public_schema(self, locale: str):
-        """Generates a JSON-encodeable schema describing this entity."""
-        return {
-            "name": self.name,
-            "caption": self.get_caption(locale),
-            "type": self.key_type.value,
-            "annotations": self.annotations,
-        }
-
 
 def get_shared_entity(shared_map: Mapping[str, T], name: str) -> T:
     try:
         return shared_map[name]
     except KeyError:
         raise EntityUsageError(name, "SharedDimension") from None
+
+
+def validate_entity_names(
+    cube: str, item: Union["AnyMeasure", "EntityUsageTraverser"], record: Dict[str, str]
+):
+    """Verifies the name of the Entity is valid and unique."""
+    if isinstance(item, AnyMeasure):
+        origin = item
+    else:
+        origin = item._entity if item._usage is None else item._usage
+    name = origin.name
+    entity_type = type(origin).__name__
+
+    if "," in name:
+        raise InvalidNameError(cube, entity_type, name)
+
+    if name in record:
+        raise DuplicatedNameError(cube, record[name], entity_type, name)
+
+    record[name] = entity_type
