@@ -7,18 +7,32 @@ import logging
 from ast import literal_eval as eval_tuple
 from collections import OrderedDict
 from pathlib import Path
-from typing import (Generator, Iterable, List, Optional, TextIO, Tuple, Type,
-                    TypeVar, Union)
+from typing import (
+    Generator,
+    Iterable,
+    List,
+    Optional,
+    TextIO,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
+)
 
-import httpx
 import immutables as immu
 from lxml import etree
 
+from tesseract_olap.common import is_numeric, numerify
+from tesseract_olap.exceptions.schema import (
+    InvalidXMLAttributeValue,
+    MissingXMLAttribute,
+    MissingXMLNode,
+)
+
 from . import models
 from .aggregators import Aggregator
 from .csv import parse_csv
 from .enums import AggregatorType, DimensionType, MemberType
-from .exceptions import InvalidXMLAttributeValue, MissingXMLAttribute, MissingXMLNode
 
 logger = logging.getLogger(__name__)
 
@@ -35,6 +49,7 @@ XMLEntity = Union[
     "XMLPropertyUsage",
     "XMLPrivateDimension",
     "XMLMeasure",
+    "XMLCalculatedMeasure",
 ]
 
 AnyXMLEntity = TypeVar("AnyXMLEntity", bound=XMLEntity)
@@ -44,14 +59,17 @@ class XMLSchema(models.Schema):
     tag = "Schema"
 
     @classmethod
-    def parse(cls, node: etree._Element):
+    def parse(cls, node: etree._Element, index: int = 0):
         """Parse a <Schema> XML node."""
+        name = _get_attr(node, "name")
+        logger.debug("Parsing node <%s name='%s' />", cls.tag, name)
+
         cube_gen = _yield_children_nodes(node, XMLCube)
         shareddim_gen = _yield_children_nodes(node, XMLSharedDimension)
         sharedtbl_gen = _yield_children_nodes(node, XMLInlineTable)
 
         return cls(
-            name=_get_attr(node, "name"),
+            name=name,
             cube_map=OrderedDict(cube_gen),
             shared_dimension_map=immu.Map(shareddim_gen),
             shared_table_map=immu.Map(sharedtbl_gen),
@@ -60,35 +78,60 @@ class XMLSchema(models.Schema):
         )
 
 
+class XMLAccessControl(models.AccessControl):
+    tag = "Access"
+
+    @classmethod
+    def parse(cls, node: etree._Element):
+        """Parse all <Access> XML node that are directly under this node."""
+
+        rules_gen = (
+            (role, _get_attr(item, "rule") == "allow")
+            for item in node.iterchildren(cls.tag)
+            for role in _get_attr(item, "roles").split(",")
+        )
+
+        return cls(
+            public=node.get("public", "true") == "true",
+            rules=immu.Map(rules_gen),
+        )
+
+
 class XMLCube(models.Cube):
     tag = "Cube"
 
     @classmethod
-    def parse(cls, node: etree._Element, _: int):
+    def parse(cls, node: etree._Element, index: int):
         """Parse a <Cube> XML node."""
-        cube_name = _get_attr(node, "name")
+        name = _get_attr(node, "name")
+        logger.debug("Parsing node <%s name='%s' />", cls.tag, name)
 
         table = _find_table_ref(node)
         if table is None:
-            raise MissingXMLNode(node.tag, cube_name, "Table")
+            raise MissingXMLNode(node.tag, name, "Table")
 
         dimension_map = OrderedDict(
             _yield_children_nodes(node, XMLPrivateDimension, XMLDimensionUsage)
         )
         if len(dimension_map) == 0:
-            raise MissingXMLNode(node.tag, cube_name, "Dimension")
+            raise MissingXMLNode(node.tag, name, "Dimension")
 
-        measure_map = OrderedDict(_yield_children_nodes(node, XMLMeasure))
+        measure_map = OrderedDict(
+            _yield_children_nodes(node, XMLMeasure, XMLCalculatedMeasure)
+        )
         if len(measure_map) == 0:
-            raise MissingXMLNode(node.tag, cube_name, "Measure")
+            raise MissingXMLNode(node.tag, name, "Measure")
 
         return cls(
-            name=cube_name,
+            name=name,
             captions=immu.Map(_yield_locale_pairs(node, "caption")),
+            acl=XMLAccessControl.parse(node),
             dimension_map=dimension_map,
             measure_map=measure_map,
             table=table,
             annotations=immu.Map(_yield_annotations(node)),
+            subset_table=node.get("subset_table", "false").lower() != "false",
+            visible=node.get("visible", "true").lower() != "false",
         )
 
 
@@ -96,12 +139,15 @@ class XMLDimensionUsage(models.DimensionUsage):
     tag = "DimensionUsage"
 
     @classmethod
-    def parse(cls, node: etree._Element, _: int):
+    def parse(cls, node: etree._Element, index: int):
         """Parse a <DimensionUsage> XML node."""
+        name = _get_attr(node, "name")
+        logger.debug("Parsing node <%s name='%s' />", cls.tag, name)
+
         return cls(
-            name=_get_attr(node, "name"),
+            name=name,
             source=_get_attr(node, "source"),
-            foreign_key=node.get("foreign_key"),
+            foreign_key=_get_attr(node, "foreign_key"),
             annotations=immu.Map(_yield_annotations(node)),
             captions=immu.Map(_yield_locale_pairs(node, "caption")),
             hierarchy_map=OrderedDict(
@@ -114,10 +160,13 @@ class XMLHierarchyUsage(models.HierarchyUsage):
     tag = "HierarchyUsage"
 
     @classmethod
-    def parse(cls, node: etree._Element, _: int):
+    def parse(cls, node: etree._Element, index: int):
         """Parse a <HierarchyUsage> XML node."""
+        name = _get_attr(node, "name")
+        logger.debug("Parsing node <%s name='%s' />", cls.tag, name)
+
         return cls(
-            name=_get_attr(node, "name"),
+            name=name,
             source=_get_attr(node, "source"),
             annotations=immu.Map(_yield_annotations(node)),
             captions=immu.Map(_yield_locale_pairs(node, "caption")),
@@ -131,10 +180,13 @@ class XMLLevelUsage(models.LevelUsage):
     tag = "LevelUsage"
 
     @classmethod
-    def parse(cls, node: etree._Element, _: int):
+    def parse(cls, node: etree._Element, index: int):
         """Parse a <LevelUsage> XML node."""
+        name = _get_attr(node, "name")
+        logger.debug("Parsing node <%s name='%s' />", cls.tag, name)
+
         return cls(
-            name=_get_attr(node, "name"),
+            name=name,
             source=_get_attr(node, "source"),
             annotations=immu.Map(_yield_annotations(node)),
             captions=immu.Map(_yield_locale_pairs(node, "caption")),
@@ -148,10 +200,13 @@ class XMLPropertyUsage(models.PropertyUsage):
     tag = "PropertyUsage"
 
     @classmethod
-    def parse(cls, node: etree._Element, _: int):
+    def parse(cls, node: etree._Element, index: int):
         """Parse a <PropertyUsage> XML node."""
+        name = _get_attr(node, "name")
+        logger.debug("Parsing node <%s name='%s' />", cls.tag, name)
+
         return cls(
-            name=_get_attr(node, "name"),
+            name=name,
             source=_get_attr(node, "source"),
             annotations=immu.Map(_yield_annotations(node)),
             captions=immu.Map(_yield_locale_pairs(node, "caption")),
@@ -162,12 +217,15 @@ class XMLTable(models.Table):
     tag = "Table"
 
     @classmethod
-    def parse(cls, node: etree._Element, _: int):
+    def parse(cls, node: etree._Element, index: int):
         """Parse a <Table> XML node."""
+        name = _get_attr(node, "name")
+        logger.debug("Parsing node <%s name='%s' />", cls.tag, name)
+
         return cls(
-            name=_get_attr(node, "name"),
+            name=name,
             schema=node.get("schema"),
-            primary_key=node.get("primary_key"),
+            primary_key=node.get("primary_key", "id"),
         )
 
 
@@ -175,9 +233,11 @@ class XMLInlineTable(models.InlineTable):
     tag = "InlineTable"
 
     @classmethod
-    def parse(cls, node: etree._Element, _: int):
+    def parse(cls, node: etree._Element, index: int):
         """Parse a <InlineTable> XML node."""
-        node_name = _get_attr(node, "name")
+        name = _get_attr(node, "name")
+        logger.debug("Parsing node <%s name='%s' />", cls.tag, name)
+
         node_format = _get_attr(node, "format")
 
         if node_format == "csv":
@@ -188,22 +248,25 @@ class XMLInlineTable(models.InlineTable):
         elif node_format.startswith("text/csv"):
             content = [] if node.text is None else node.text.strip().splitlines()
             headers, *rows = parse_csv(content, mimetype=node_format)
+            headers = tuple(str(item) for item in headers)
         else:
-            raise InvalidXMLAttributeValue(node.tag, node_name, "format", node_format)
+            raise InvalidXMLAttributeValue(node.tag, name, "format", node_format)
 
         return cls(
-            name=node_name,
+            name=name,
             headers=headers,
-            types=tuple(MemberType.from_values(column) for column in zip(*rows)),
+            types=cls.infer_types(rows),
             rows=rows,
         )
 
     @staticmethod
-    def parse_tuples(node: etree._Element):
+    def parse_tuples(
+        node: etree._Element,
+    ) -> Tuple[Tuple[str, ...], List[Tuple[Union[str, float], ...]]]:
         """Parse the child nodes from an InlineTable with `tuples` format."""
         try:
             # try parsing literal tuples with ast.eval()
-            children: List[Tuple[str, ...]] = [
+            children: List[Tuple[Union[str, float], ...]] = [
                 eval_tuple(line)
                 for line in (item.text for item in node.iterchildren("Row"))
                 if line
@@ -219,23 +282,34 @@ class XMLInlineTable(models.InlineTable):
         # at least 2 rows must be present: a header list and a data row
         if len(children) < 2:
             raise MissingXMLNode(node.tag, _get_attr(node, "name"), "Row")
-        return children[0], children[1:]
+        headers = tuple(str(item) for item in children[0])
+        return headers, children[1:]
 
 
 class XMLSharedDimension(models.Dimension):
     tag = "SharedDimension"
 
     @classmethod
-    def parse(cls, node: etree._Element, _: int):
+    def parse(cls, node: etree._Element, index: int):
         """Parse a Shared <Dimension> XML node."""
+        name = _get_attr(node, "name")
+        logger.debug("Parsing node <%s name='%s' />", cls.tag, name)
+
         dim_type = node.get("type")
 
+        hierarchy_map = OrderedDict(_yield_children_nodes(node, XMLHierarchy))
+
+        default_hierarchy = node.get("default_hierarchy")
+        if default_hierarchy not in hierarchy_map:
+            default_hierarchy = next(iter(hierarchy_map.keys()))
+
         return cls(
-            name=_get_attr(node, "name"),
+            name=name,
             captions=immu.Map(_yield_locale_pairs(node, "caption")),
+            default_hierarchy=default_hierarchy,
             dim_type=DimensionType.from_str(dim_type),
             foreign_key=node.get("foreign_key"),
-            hierarchy_map=OrderedDict(_yield_children_nodes(node, XMLHierarchy)),
+            hierarchy_map=hierarchy_map,
             annotations=immu.Map(_yield_annotations(node)),
         )
 
@@ -259,20 +333,21 @@ class XMLHierarchy(models.Hierarchy):
     tag = "Hierarchy"
 
     @classmethod
-    def parse(cls, node: etree._Element, _: int):
+    def parse(cls, node: etree._Element, index: int):
         """Parse a <Hierarchy> XML node."""
-        node_name = _get_attr(node, "name")
+        name = _get_attr(node, "name")
+        logger.debug("Parsing node <%s name='%s' />", cls.tag, name)
 
         level_map = OrderedDict(_yield_children_nodes(node, XMLLevel))
         if len(level_map) == 0:
-            raise MissingXMLNode(node.tag, node_name, "Level")
+            raise MissingXMLNode(node.tag, name, "Level")
 
         default_pk = ""
         for item in level_map.values():
             default_pk = item.key_column
 
         return cls(
-            name=node_name,
+            name=name,
             primary_key=node.get("primary_key", default_pk),
             captions=immu.Map(_yield_locale_pairs(node, "caption")),
             table=_find_table_ref(node),
@@ -284,7 +359,7 @@ class XMLHierarchy(models.Hierarchy):
     @staticmethod
     def _parse_default_member(node: etree._Element):
         items: List[str] = node.get("default_member", "").split(".", maxsplit=1)
-        return tuple(items) if len(items) == 2 else None
+        return (items[0], items[1]) if len(items) == 2 else None
 
 
 class XMLLevel(models.Level):
@@ -293,10 +368,13 @@ class XMLLevel(models.Level):
     @classmethod
     def parse(cls, node: etree._Element, index: int):
         """Parse a <Level> XML node."""
+        name = _get_attr(node, "name")
+        logger.debug("Parsing node <%s name='%s' />", cls.tag, name)
+
         key_type = node.get("key_type")
 
         return cls(
-            name=_get_attr(node, "name"),
+            name=name,
             depth=index + 1,
             key_column=_get_attr(node, "key_column"),
             key_type=MemberType.from_str(key_type),
@@ -311,8 +389,11 @@ class XMLProperty(models.Property):
     tag = "Property"
 
     @classmethod
-    def parse(cls, node: etree._Element, _: int):
+    def parse(cls, node: etree._Element, index: int):
         """Parse a <Property> XML node."""
+        name = _get_attr(node, "name")
+        logger.debug("Parsing node <%s name='%s' />", cls.tag, name)
+
         key_type = node.get("key_type")
 
         keycol_map = immu.Map(_yield_locale_pairs(node, "key_column"))
@@ -320,7 +401,7 @@ class XMLProperty(models.Property):
             raise MissingXMLAttribute(node.tag, "key_column")
 
         return cls(
-            name=_get_attr(node, "name"),
+            name=name,
             annotations=immu.Map(_yield_annotations(node)),
             captions=immu.Map(_yield_locale_pairs(node, "caption")),
             key_column_map=keycol_map,
@@ -332,25 +413,18 @@ class XMLMeasure(models.Measure):
     tag = "Measure"
 
     @classmethod
-    def parse(cls, node: etree._Element, _: int):
+    def parse(cls, node: etree._Element, index: int):
         """Parse a <Measure> XML node."""
-        submeasures = (
-            (item.name, item)
-            for item in (
-                models.SubMeasure(
-                    name=_get_attr(item, "name"),
-                    aggregator=cls._get_aggregator(item),
-                )
-                for item in node.iterchildren("Submeasure")
-            )
-        )
+        name = _get_attr(node, "name")
+        logger.debug("Parsing node <%s name='%s' />", cls.tag, name)
+
         return cls(
-            name=_get_attr(node, "name"),
+            name=name,
             key_column=_get_attr(node, "key_column"),
             aggregator=cls._get_aggregator(node),
             annotations=immu.Map(_yield_annotations(node)),
             captions=immu.Map(_yield_locale_pairs(node, "caption")),
-            submeasures=immu.Map(submeasures),
+            submeasures=immu.Map(_yield_children_nodes(node, cls)),
         )
 
     @staticmethod
@@ -379,11 +453,33 @@ class XMLMeasure(models.Measure):
             raise InvalidXMLAttributeValue(node.tag, node_name, attr, value)
         else:
             agg_cls = Aggregator.from_enum(agg_type)
-            return agg_cls.new(node.attrib)
+            agg_args = {
+                str(k).replace("-", "_"): numerify(v) if is_numeric(v) else str(v)
+                for k, v in node.attrib.items()
+            }
+            return agg_cls.new(agg_args)
+
+
+class XMLCalculatedMeasure(models.CalculatedMeasure):
+    tag = "CalculatedMeasure"
+
+    @classmethod
+    def parse(cls, node: etree._Element, index: int):
+        """Parse a <CalculatedMeasure> XML node."""
+        name = _get_attr(node, "name")
+        logger.debug("Parsing node <%s name='%s' />", cls.tag, name)
+
+        return cls(
+            name=name,
+            formula=cls._parse_formula(_get_attr(node, "formula")),
+            annotations=immu.Map(_yield_annotations(node)),
+            captions=immu.Map(_yield_locale_pairs(node, "caption")),
+        )
 
 
 def _find_table_ref(node: etree._Element):
-    table_node = next(node.iterchildren("InlineTable", "Table", "TableUsage"), None)
+    gen_tables = (item for item in node.iterchildren("InlineTable", "Table", "TableUsage"))
+    table_node = next(gen_tables, None)
 
     if table_node is None:
         return None
@@ -398,15 +494,20 @@ def _find_table_ref(node: etree._Element):
 
 
 def _get_attr(node: etree._Element, attr: str) -> str:
+    """Retrieves an attribute from a node.
+
+    If the attribute is not present, raises :class:`MissingXMLAttribute`.
+    """
     try:
         value = node.attrib[attr]
-    except KeyError:
-        raise MissingXMLAttribute(node.tag, attr)
+    except KeyError as exc:
+        raise MissingXMLAttribute(node.tag, attr) from exc
     else:
-        return value
+        return str(value)
 
 
 def _yield_annotations(node: etree._Element) -> Iterable[Tuple[str, Optional[str]]]:
+    """Yields a pair of (name, value) for each Annotation in the node."""
     return (
         (_get_attr(item, "name"), item.text)
         for item in node.iterchildren("Annotation")
@@ -441,77 +542,24 @@ def _yield_locale_pairs(node: etree._Element,
             yield (_get_attr(child_node, "locale"), child_value)
 
 
-def _parse_pathlib_path(path: Path, parser: etree.XMLParser) -> XMLSchema:
-    """Parses an XML schema from the content of a specific pathlib.Path instance.
-
-    This function is able to parse a single file, or all files in a directory.
-    """
-    if not path.exists():
-        raise FileNotFoundError("XML source: Path {} does not exist".format(path))
-
-    def read_xml_path(path: Path) -> XMLSchema:
-        with path.open("r") as fileio:
-            tree = etree.parse(fileio, parser)
-        root = tree.getroot()
-        return XMLSchema.parse(root)
-
-    # If path is directory, load all XMLs and combine them
-    if path.is_dir():
-        schemas = (read_xml_path(item)
-                  for item in path.glob("**/*.xml") if item.is_file())
-        return XMLSchema.join(*schemas)
-
-    # If path is file, load and parse it
-    if path.is_file():
-        return read_xml_path(path)
-
-    raise ValueError("XML source: Path can't be interpreted")
-
-
 def parse_xml_schema(source: Union[str, Path, TextIO]) -> XMLSchema:
     """Attempts to parse an object into a XMLSchema.
 
     This function accepts:
-    - An URL (as a :class:`str`)
-    - A local path (as a :class:`str` or :class:`pathlib.Path`) to a XML file,
-      or a directory containing XML files
     - A raw XML :class:`str`
+    - A local path (as a :class:`pathlib.Path`) to a XML file
     - A not-binary read-only :class:`TextIO` instance for a file-like object
     """
     parser = etree.XMLParser(encoding="utf-8",
                              remove_blank_text=True,
                              remove_comments=True)
 
-    # if argument is string...
+    # if argument is str, is assumed to be a raw XML string
     if isinstance(source, str):
-        source = source.strip()
+        root = etree.fromstring(source, parser)
+        return XMLSchema.parse(root)
 
-        # Check if argument is a URL and fetch the file
-        if source.startswith(("http://", "https://", "ftp://")):
-            response = httpx.get(source)
-            response.raise_for_status()
-
-            root = etree.fromstring(response.text, parser)
-            return XMLSchema.parse(root)
-
-        # Check if argument is a path and load the file(s)
-        elif source.endswith((".xml", "/", "\\")):
-            path = Path(source).resolve()
-            return _parse_pathlib_path(path, parser)
-
-        # Check if argument is a raw XML string
-        elif source.startswith("<Schema "):
-            root = etree.fromstring(source, parser)
-            return XMLSchema.parse(root)
-
-        raise ValueError("XML source: Value can't be recognized")
-
-    # if argument is a pathlib.Path, open it and parse contents
-    elif isinstance(source, Path):
-        path = source.resolve()
-        return _parse_pathlib_path(path, parser)
-
-    # if argument is not a string, attempt to use it like a file-like object
+    # if argument is pathlib.Path or TextIO, open it and parse contents
     else:
         tree = etree.parse(source, parser)
         root = tree.getroot()
